/*
Copyright (c) 2015  V.Ayllón.  All rights reserved.
 
 TwinTeethMC is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 TwinTeethMC, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with TwinTeethMC so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

import java.util.*;


public class Simulation {
  int[] DRAW_WINDOW_SIZE = {
    DRAW_WINDOW_COORDS[2]-DRAW_WINDOW_COORDS[0], DRAW_WINDOW_COORDS[3]-DRAW_WINDOW_COORDS[1], 85
  };

  private int COORDX_ORIGIN = 0;
  private int COORDY_ORIGIN = 0;
  private int COORDZ_ORIGIN = 0; 


  private float scaler = 0.;
  private float nmx, nmy, nmz= 0.;
  private float rx, ry, rz = 0.;
  private float ejxOffset, xOffset = 0.0; 
  private float ejyOffset, yOffset = 0.0; 

  private static final int MAX_CMD_SIZE = 255;
  private static final int MAX_FLOAT_SIZE = 80;
  private char cmdbuffer[] = new char[MAX_CMD_SIZE];
  private int buffinx = 0;
  private boolean comment_mode = false;
  private static final int X_AXIS = 0;
  private static final int Y_AXIS = 1;
  private static final int Z_AXIS = 2;
  private static final int E_AXIS = 3;
  private static final int NUM_AXIS = 4;
  private static final int  I_COORD  = 0;
  private static final int  J_COORD  = 1;
  private static final float DRAW_SCALE = 4;
  private static final float DRILL_RADIUS = .2*DRAW_SCALE;
  private static final float PASTE_WIDTH = .2*DRAW_SCALE;
  private static final float FIXTURE_WIDTH = 70;
  private static final float FIXTURE_HEIGHT = 80;
  private static final float FIXTURE_DEEP = 5;
  private static final float FIXTURE_ZPOSITION = -4.1;

  private char[] axisCodes = {
    'X', 'Y', 'Z', 'E'
  };

  public boolean relativeMode = false;  //Determines Absolute or Relative Coordinates

  private float[] offset =  new float[NUM_AXIS];
  private float[] currentPosition = new float[NUM_AXIS];
  private float[] destination = new float[NUM_AXIS];
  private float[] drawCurrentPosition = new float[NUM_AXIS];
  private float[] drawDestination = new float[NUM_AXIS];
  private float dx;
  private float dy;
  private float dz;
  private float de;

  private float feedrate = 0;
  private float feedmultiply = 1;
  private boolean fdState = false;

  private static final float MM_PER_ARC_SEGMENT_SMALL = .1;
  private static final float SMALL_CIRCLE_RADIUS = 2.5;
  private static final float MM_PER_ARC_SEGMENT_LARGE = 1;

  private static final int N_ARC_CORRECTION = 25;



  public PGraphics mypg;

  ///// INNER classes


  /***
   ** class Point
   **/
  private class Point {
    float x;
    float y;
    float z;
    float e;

    Point(float px, float py, float pz, float pe) {
      x = px;
      y = py;
      z = pz;
      e = pe;
    }
  }
  /***
   ** class Line
   **/
  private class Line {
    Point current;
    Point destination;
    boolean feedState;

    Line(Point c, Point d, boolean fs) {
      current = c;
      destination = d;
      feedState = fs;
    }
  }
  /***
   ** class Slice
   **/
  private class Slice {
    private ArrayList<Line> slice = new ArrayList<Line>(); 

    Slice() {
    }

    Slice(ArrayList<Line> s) {
      slice = s;
    }

    // Draw the slice
    public void draw(color c) {
      for (int i = 0; i < slice.size (); i++) {
        Line l = slice.get(i);
        if (l.feedState) {
          mypg.stroke(c);
          mypg.line(l.current.x, l.current.y, l.current.z, l.destination.x, l.destination.y, l.destination.z);
        }
      }
    }

    // Add a line
    public void add(Line l) {
      slice.add(l);
    }
  }
  /***
   ** class 3Dobject
   **/
  private class Object3D {

    private ArrayList<Slice> obj = new ArrayList<Slice>();
    private int currentSlice;
    private color objColor;
    /// Constructors

    Object3D(color c) {
      currentSlice = 0;
      objColor = c;
    }

    Object3D(ArrayList<Slice> o, color c) {
      currentSlice = 0;
      obj = o;
      objColor = c;
    }
    // Add an slice

    public void add(Slice s) {
      obj.add(s);
    }

    // Draw the object

    public void drawAll() {
      for (int i = 0; i < obj.size (); i++) {
        obj.get(i).draw(objColor);
      }
    }

    // Draw a slice

    public void drawSlice(int sliceNum, color c) {
      if (obj.size() > sliceNum)
        obj.get(sliceNum).draw(c);
    }

    // Clear de object

    public void clear() {
      obj.clear();
      currentSlice = 0;
    }

    public void drawToCurrent(color c) {
      if (obj.size() > 0 && currentSlice < obj.size()) {
        for (int i = 0; i <= currentSlice; i++) {
          obj.get(i).draw(c);
        }
      }
    }

    public void nextSlice() {
      if (currentSlice < obj.size())
        currentSlice++;
    }

    public int getNumSlices() {
      return obj.size();
    }

    public int getCurrentSlice() {
      return currentSlice;
    }

    public void resetCurrentSlice() {
      currentSlice = 0;
    }
  }

  Slice currentSlice = new Slice();
  Object3D obj3D = new Object3D(myColor);



  ////////// Simulation Constructors & Methods

  Simulation() {
    mypg = createGraphics(DRAW_WINDOW_SIZE[0], DRAW_WINDOW_SIZE[1], P3D);
    mypg.smooth();
    initObjectPosition();
  }

  Simulation(PGraphics pgo) {
    mypg = pgo;
    mypg.smooth();
    initObjectPosition();
  }


  /** 
   ** updateGraphics() -  Updates the image
   **/
  public void updateGraphics()
  {
    image(mypg, DRAW_WINDOW_COORDS[0], DRAW_WINDOW_COORDS[1], DRAW_WINDOW_COORDS[2], DRAW_WINDOW_COORDS[3]);
  }
  /** 
   ** refresh() - Redraw the simulation
   **/
  public void refresh()
  {
    mypg.beginDraw();
    if (printState != PRINTING_STATE)
      mypg.background(G3DDBACKGROUND);  // Clean the window
    mypg.translate((DRAW_WINDOW_SIZE[0]/2) + nmx, (DRAW_WINDOW_SIZE[1]/2) + nmy, (DRAW_WINDOW_SIZE[2]/2) + nmz);
    mypg.scale(scaler);
    mypg.rotateX(rx);
    mypg.rotateY(ry);
    mypg.rotateZ(rz);

    if (activeTool == LASER_TOOL && !(laserModeCtrl.getState(0) || laserModeCtrl.getState(1))) { // Rastering
      if (printState == PRINTING_STATE) {
        raster.drawToCurrentLine(blueRay);
      }
      raster.updateGraphics(mypg);
    } else if (activeTool == EXTRUDER_TOOL || (activeTool == LASER_TOOL && laserModeCtrl.getState(1))) {  //3D
      drawFixtureBed();
      if (printState != PRINTING_STATE) {
        drawAll();
      } else {
        if (sim.isNewSlice(buffer)) {   // If new slice print it
          drawToCurrentSlice(blueRay);
          nextSlice();
        }
      }
    } else { // Rest of tools simulate GCode
      drawFixtureBed();
      for (int i = 0; i < fileRow.length; i++) {
        if (rowNum == 0) {  // not started print
          myColor = grayRay;
        } else {              // Printing or paused
          if (i < rowNum)
            myColor = blueRay;
          else
            myColor = grayRay;
        }
        simulateCommand(fileRow[i]);
      }
      homeAllAxis();
    }
    // Redraw

    mypg.endDraw();
  }


  /** 
   ** init() - Init simulation
   **/
  public void init()
  {
    obj3D.resetCurrentSlice();
    homeAllAxis();
  }

  // Clear de object

  public void clear() {
    obj3D.clear();
    currentSlice = null;
  }

  /** 
   ** loadObject() - Slice a  3D gcode file
   **/
  public void loadObject(String gcodeRow[])
  {
    if (activeTool == EXTRUDER_TOOL || (activeTool == LASER_TOOL && laserModeCtrl.getState(1))) {
      clear();
      currentSlice = new Slice();
      for (int i = 0; i < gcodeRow.length; i++) {
        getCommand(gcodeRow[i]);
        buffinx = 0;
        process3DCommands(true);
      }
    }
    //printlnEcho("Object loaded:"+ obj3D.getNumSlices() + " layers\n");
  }
  /** 
   ** drawAll() - Print a 3D object previosly loaded
   **/
  public void drawAll()
  {
    obj3D.drawAll();
  }
  /** 
   ** drawToCurrentSlice() - Change the color till current slice
   **/
  public void drawToCurrentSlice(color c)
  {
    obj3D.drawToCurrent(c);
    printlnMsg("Printing layer: " + obj3D.getCurrentSlice()+"/"+obj3D.getNumSlices());
  }
  /** 
   ** nextSlice() - Increment the current slice
   **/
  public void nextSlice() {
    obj3D.nextSlice();
  }
  /** 
   ** isNewSlice() - Checks if printer changed to a new slice
   **/
  public boolean isNewSlice(String buff)
  {
    getCommand(buff);
    buffinx = 0;
    return process3DCommands(false);
  }

  /** 
   ** preProcessCommand() - 12-11-2015 - VAF
   ** Utility method which pre-process the gcode command and
   ** add X,Y offsets to coordinates if needed. 
   ** Return the command string modified.
   **/
  public String preProcessCommand(String gcodeRow) {

    int tmpInx = 0;
    String newCMD = new String();
    int buffLen = 0;

    // Parse the command
    getCommand(gcodeRow);
    buffLen = buffinx;
    buffinx = 0;

    try {
      if (xOffsetCtrl.getValue() != 0f) {
        if (codeSeen(axisCodes[X_AXIS]))
        {
          newCMD += String.copyValueOf(cmdbuffer, 0, buffinx);
          newCMD += String.format("%2.4f ", (codeValue() + xOffsetCtrl.getValue()));
          tmpInx = buffinx;
        } else {
          buffinx = 0;
        }
      }

      if (yOffsetCtrl.getValue() != 0f) {
        if (codeSeen(axisCodes[Y_AXIS])) 
        {
          newCMD += String.copyValueOf(cmdbuffer, tmpInx, buffinx-tmpInx);
          newCMD += String.format("%2.4f ", (codeValue() + yOffsetCtrl.getValue()));
          tmpInx = buffinx;
        }
      }
      if (buffLen-tmpInx > 0)
        newCMD += String.copyValueOf(cmdbuffer, tmpInx, buffLen-tmpInx);
    } 
    catch (NumberFormatException nfe) {
      System.out.println("Error converting g-code file " + nfe.toString());
    }

    return newCMD;
  }


  /** 
   ** simulateCommand() - Simulate the execution of a command
   **/
  public void simulateCommand(String buff)
  {
    getCommand(buff);
    buffinx = 0;
    processCommands();
  }

  /** 
   ** getCommand() - Reads the buffer character-by-character, identifies comments and filter
   **/
  void getCommand(String s)
  {
    char c;
    int count = 0;
    buffinx = 0;
    while (count < s.length ()) {
      c = s.charAt(count++);
      if ((c < ' ') || (c > 'z')
        || ((c > 'Z') && (c < 'a')))
        continue;
      if (comment_mode) { // comments 
        if (c == ')')  comment_mode = false;
        continue;
      } else { // not comment
        if (c == ';' || c == '(') {
          comment_mode = true;
        } else {
          // good characters
          cmdbuffer[buffinx++] = c;
        }
      }
    }
    comment_mode = false;
    cmdbuffer[buffinx] = '\0';
    return;
  }


  /** 
   ** processCommands() - Reads the buffer character-by-character, identifies comments and filter
   **/
  void processCommands()
  {
    float z_up = 0.;
    float z_depth  = 0.;

    try {
      if (codeSeen('G'))
      {
        switch((int)codeValue())
        {
        case 0: // G0 
        case 1: // G1
          getCoordinates(); 
          prepareMove();
          lineMove();
          break;
        case 2: // G2  - CW ARC
          getArcCoordinates();
          prepareArcMove(true);
          break;
        case 3: // G3  - CCW ARC
          getArcCoordinates();
          prepareArcMove(false);
          break;
        case 28: // G28
          homeAllAxis();
          break;
        case 81: // G81 Drill command
          getCoordinates(); 
          z_depth = destination[Z_AXIS];
          if (codeSeen('R')) {
            if (relativeMode)
              z_up = codeValue() + currentPosition[Z_AXIS];
            else
              z_up = codeValue();
          }

          destination[Z_AXIS] = z_up;
          prepareMove();
          lineMove();

          destination[Z_AXIS] = z_depth;
          prepareMove();
          lineMove();

          destination[Z_AXIS] = z_up;
          prepareMove();
          lineMove();

          break;
        case 90: // G90
          relativeMode = false;
          break;
        case 91: // G91
          relativeMode = true;
          break;
        case 92: // G92 set offset
          if (codeSeen(axisCodes[E_AXIS])) {
            currentPosition[E_AXIS] = codeValue();
          }
          break;
        } // Switch
      } else {
        // rest of commands ignore
      }
    } 
    catch (NumberFormatException nfe) {
      nfe.printStackTrace();
    }
  }



  /** 
   ** process3DCommands() - Reads the buffer character-by-character, identifies comments and filter, execute commands.
   **/
  boolean process3DCommands(boolean scan)
  {
    float z_up = 0.;
    float z_depth  = 0.;
    try {
      if (codeSeen('G'))
      {
        switch((int)codeValue())
        {
        case 0: // G0 
        case 1: // G1
          getCoordinates(); 
          prepareMove();
          if (scan) {
            scan3DMove();
          } else {
            for (int i=0; i < NUM_AXIS; i++) {
              currentPosition[i] = destination[i];
            }
            if (dz > 0)
              return true;
          } 
          break;
        case 28: // G28
          homeAllAxis();
          break;
        case 90: // G90
          relativeMode = false;
          break;
        case 91: // G91
          relativeMode = true;
          break;
        case 92: // G92 set offset
          if (codeSeen(axisCodes[E_AXIS])) {
            currentPosition[E_AXIS] = codeValue();
          }
          break;
        } // Switch
      } else {
        // rest of commands ignore
      }
    } 
    catch (NumberFormatException nfe) {
      nfe.printStackTrace();
    }
    return false;
  }

  /** 
   ** getCoordinates() - Get linear coordinates
   **/
  void getCoordinates() throws NumberFormatException 
  {
    for (short i=0; i < NUM_AXIS; i++) {
      if (codeSeen(axisCodes[i]))
      {
        if (relativeMode)
          destination[i] = codeValue() + currentPosition[i] ;
        else
          destination[i] = codeValue();
        // 12-11-2015 - VAF - Added x,y offset to coordinates
        // becasue Eagle light does not support negative ones
        if (i == X_AXIS)
          destination[i] += xOffsetCtrl.getValue();
        if (i == Y_AXIS)
          destination[i] += yOffsetCtrl.getValue();
      } else {
        buffinx = 0;
        destination[i] = currentPosition[i]; //Are these else lines really needed?
      }
    }
  }
  /** 
   ** getArcCoordinates() - Get arc coordinates
   **/
  void getArcCoordinates() throws NumberFormatException 
  {

    getCoordinates();

    if (codeSeen('I')) {
      offset[I_COORD] = codeValue();
    } else {
      offset[I_COORD] = 0.0;
    }
    if (codeSeen('J')) {
      offset[J_COORD] = codeValue();
    } else {
      offset[J_COORD] = 0.0;
    }
  }

  //////////////////////// MOTION CONTROL ////////////////////////


  /**
   ** homeAllAxis() - Home all axis
   **/
  public void homeAllAxis() {
    fdState = OFF;
    relativeMode = false;
    currentPosition[X_AXIS] = 0.;
    currentPosition[Y_AXIS]  = 0.;
    currentPosition[Z_AXIS] = 0.;  
    destination[X_AXIS] = 0.;
    destination[Y_AXIS] = 0.;
    destination[Z_AXIS] = 0.;
    drawCurrentPosition[X_AXIS] = COORDX_ORIGIN;
    drawCurrentPosition[Y_AXIS] = COORDY_ORIGIN;
    drawCurrentPosition[Z_AXIS] = COORDZ_ORIGIN;
    drawDestination[X_AXIS] = 0.;
    drawDestination[Y_AXIS] = 0.;
    drawDestination[Z_AXIS] = 0.;
    dx = dy = dz = de = 0.;
  }

  /**
   * prepareMove() - Prepare the movement
   **/
  public void prepareMove()
  {

    if (destination[Z_AXIS] < 0)
      fdState = ON;
    else 
    fdState = OFF;
    //System.out.print("X " + currentPosition[X_AXIS]);
    //System.out.print(" Y " + currentPosition[Y_AXIS]);
    //System.out.println(" Z " + currentPosition[Z_AXIS]);

    // Update draw positions
    drawCurrentPosition[X_AXIS] =  COORDX_ORIGIN + currentPosition[X_AXIS]*DRAW_SCALE;
    drawCurrentPosition[Y_AXIS] =  COORDY_ORIGIN - currentPosition[Y_AXIS]*DRAW_SCALE;  // Flip Y
    drawCurrentPosition[Z_AXIS] =  COORDZ_ORIGIN + currentPosition[Z_AXIS]*DRAW_SCALE;

    drawDestination[X_AXIS] = COORDX_ORIGIN + destination[X_AXIS]*DRAW_SCALE;
    drawDestination[Y_AXIS] = COORDY_ORIGIN - destination[Y_AXIS]*DRAW_SCALE; // Flip Y
    drawDestination[Z_AXIS] = COORDZ_ORIGIN + destination[Z_AXIS]*DRAW_SCALE;

    dx = abs(drawDestination[X_AXIS] -  drawCurrentPosition[X_AXIS]);
    dy = abs(drawDestination[Y_AXIS] -  drawCurrentPosition[Y_AXIS]);
    dz = abs(drawDestination[Z_AXIS] -  drawCurrentPosition[Z_AXIS]);
    de = abs(destination[E_AXIS] -  currentPosition[E_AXIS]);
  }
  /**
   * lineMove() - Paint a line
   **/
  void lineMove() {
    //println("From x:" + drawCurrentPosition[X_AXIS] +  " y:" + drawCurrentPosition[Y_AXIS] +" to x:"+ drawDestination[X_AXIS] + " y:" + drawDestination[Y_AXIS]);
    // Check active tool

    switch(activeTool) {
    case 0:  // Pen tool
    case 2:  // Laser tool Gcode
      if (max(dx, dy) > 0) {
        if (fdState) {
          mypg.stroke(myColor);
          mypg.line(drawCurrentPosition[X_AXIS], drawCurrentPosition[Y_AXIS], drawDestination[X_AXIS], drawDestination[Y_AXIS]);
        }
      }
      break;
    case 1:  // Drill
      if (fdState) {
        mypg.fill(myColor);
        mypg.ellipse(drawCurrentPosition[X_AXIS], drawCurrentPosition[Y_AXIS], DRILL_RADIUS*2, DRILL_RADIUS*2);
      } else {
        if (max(dx, dy, dz) > 0) {
          mypg.stroke(myColor);
          mypg.line(drawCurrentPosition[X_AXIS], drawCurrentPosition[Y_AXIS], drawCurrentPosition[Z_AXIS], drawDestination[X_AXIS], drawDestination[Y_AXIS], drawDestination[Z_AXIS]);
        }
      }
      break;
    case 3:  // 3D printer tool
      if (de > 0) {
        mypg.stroke(myColor);
        mypg.line(drawCurrentPosition[X_AXIS], drawCurrentPosition[Y_AXIS], drawCurrentPosition[Z_AXIS], drawDestination[X_AXIS], drawDestination[Y_AXIS], drawDestination[Z_AXIS]);
      }
      break;
    case 4:  // Solder paste dispenser
      //System.out.println("dx: " + dx + " dy: " + dy + " dz:" + dz + " de: " + de + " e: " + destination[E_AXIS]);
      if (de > 0 && destination[E_AXIS] > 0) {
        mypg.rectMode(CENTER);  // Set rectMode to CENTER
        mypg.fill(myColor); 
        mypg.rect(drawCurrentPosition[X_AXIS], drawCurrentPosition[Y_AXIS], PASTE_WIDTH*de, PASTE_WIDTH*de);
      } 
      if (max(dx, dy, dz) > 0) {
        mypg.stroke(myColor);
        mypg.line(drawCurrentPosition[X_AXIS], drawCurrentPosition[Y_AXIS], drawCurrentPosition[Z_AXIS], drawDestination[X_AXIS], drawDestination[Y_AXIS], drawDestination[Z_AXIS]);
      }
      break;
    default:
      break;
    } // switch

    // Update position

    for (int i=0; i < NUM_AXIS; i++) {
      currentPosition[i] = destination[i];
    }
  }
  /**
   * drawFixtureBed() - Draw the fixture plate
   **/
  void drawFixtureBed() {
    float scale = 0;
    scale = DRAW_SCALE;
    mypg.pushMatrix();
    mypg.translate(0, 0, FIXTURE_ZPOSITION*scale);
    mypg.noFill();
    mypg.stroke(fixtureColor);
    mypg.rectMode(CENTER);  // Set rectMode to CENTER
    mypg.box(FIXTURE_WIDTH*scale, FIXTURE_HEIGHT*scale, FIXTURE_DEEP*scale);
    mypg.popMatrix();
  }

  /**
   * scan3DMove() - Scan a 3D object movement
   **/
  void scan3DMove() {

    if (currentPosition[Z_AXIS] != 0 && dz > 0) {
      obj3D.add(currentSlice);
      currentSlice = new Slice();
    } else {
      fdState = (de > 0) ? true : false;
      currentSlice.add(new Line(new Point(drawCurrentPosition[X_AXIS], drawCurrentPosition[Y_AXIS], drawCurrentPosition[Z_AXIS], currentPosition[E_AXIS]), 
        new Point(drawDestination[X_AXIS], drawDestination[Y_AXIS], drawDestination[Z_AXIS], destination[E_AXIS]), fdState));
    }

    for (int i=0; i < NUM_AXIS; i++) {
      currentPosition[i] = destination[i];
    }
  }


  /**
   * prepareArcMove() - Prepare the movement of an arc command
   **/
  public void prepareArcMove(boolean isClockwise) {

    float radius = hypot(offset[I_COORD], offset[J_COORD]); // Compute radius

    // Trace the arc

    if (!mc_arc(currentPosition, destination, offset, X_AXIS, Y_AXIS, feedrate*feedmultiply/60/100.0, radius, isClockwise))
      return;
    // If no error update position
    for (int i=0; i < NUM_AXIS; i++) {
      currentPosition[i] = destination[i];
    }
  }

  // The arc is approximated by generating a huge number of tiny, linear segments. The length of each 
  // segment is configured in settings.mm_per_arc_segment.  
  boolean mc_arc(float[] position, float[] target, float[] offset, int axis_0, int axis_1, float feed_rate, float radius, boolean isclockwise)
  {      

    float org_targetx = target[axis_0];
    float org_targety = target[axis_1];
    //   int acceleration_manager_was_enabled = plan_is_acceleration_manager_enabled();
    //   plan_set_acceleration_manager_enabled(false); // disable acceleration management for the duration of the arc
    float center_axis0 = position[axis_0] + offset[axis_0];
    float center_axis1 = position[axis_1] + offset[axis_1];
    float r_axis0 = -offset[axis_0];  // Radius vector from center to current location
    float r_axis1 = -offset[axis_1];
    float rt_axis0 = target[axis_0] - center_axis0;
    float rt_axis1 = target[axis_1] - center_axis1;
    int segments;
    float angular_travel;

    // CCW angle between position and target from circle center. Only one atan2() trig computation required.
    angular_travel = atan2(r_axis0*rt_axis1-r_axis1*rt_axis0, r_axis0*rt_axis0+r_axis1*rt_axis1);

    if (isclockwise) { // Correct atan2 output per direction
      if (angular_travel >= 0) { 
        angular_travel -= 2*PI;
      }
    } else {
      if (angular_travel <= 0) { 
        angular_travel += 2*PI;
      }
    }

    float millimeters_of_travel = hypot(angular_travel*radius, 0.0);     

    if (millimeters_of_travel < 0.001) { 
      return false;
    }

    if (radius < SMALL_CIRCLE_RADIUS)
      segments = floor(millimeters_of_travel/MM_PER_ARC_SEGMENT_SMALL);
    else
      segments = floor(millimeters_of_travel/MM_PER_ARC_SEGMENT_LARGE);

    if (segments == 0) segments = 1;


    float theta_per_segment = angular_travel/segments;

    // Vector rotation matrix values
    float cos_T = 1-0.5*theta_per_segment*theta_per_segment; // Small angle approximation
    float sin_T = theta_per_segment;

    float[] arc_target = new float[2];
    float sin_Ti;
    float cos_Ti;
    float r_axisi;
    int i;
    int count = 0;

    for (i = 1; i<segments; i++) { // Increment (segments-1)

      if (count < N_ARC_CORRECTION) {
        // Apply vector rotation matrix 
        r_axisi = r_axis0*sin_T + r_axis1*cos_T;
        r_axis0 = r_axis0*cos_T - r_axis1*sin_T;
        r_axis1 = r_axisi;
        count++;
      } else {
        // Arc correction to radius vector. Computed only every N_ARC_CORRECTION increments.
        // Compute exact location by applying transformation matrix from initial radius vector(=-offset).
        cos_Ti = cos(i*theta_per_segment);
        sin_Ti = sin(i*theta_per_segment);
        r_axis0 = -offset[axis_0]*cos_Ti + offset[axis_1]*sin_Ti;
        r_axis1 = -offset[axis_0]*sin_Ti - offset[axis_1]*cos_Ti;
        count = 0;
      }

      // Update arc_target location
      arc_target[axis_0] = center_axis0 + r_axis0;
      arc_target[axis_1] = center_axis1 + r_axis1;
      //    clamToSoftwareEnstops(arc_target);
      /**if (fdState) {
       mypg.stroke(myColor);  // Blue ray
       mypg.line(currentx, currenty, arc_target[axis_0], arc_target[axis_1]);
       
       }**/
      destination[X_AXIS]= arc_target[axis_0];
      destination[Y_AXIS]= arc_target[axis_1];
      prepareMove();
      lineMove();
    }
    /** Ensure last segment arrives at target location.
     if (fdState) {
     mypg.stroke(myColor);  // Blue ray
     mypg.line(currentx, currenty, target[axis_0], target[axis_1]);
     }
     **/
    destination[X_AXIS]= org_targetx;
    destination[Y_AXIS]= org_targety;
    prepareMove();
    lineMove();
    return true;
  }
  /**
   * clampToSofwareEndstops() 
   **/
  public  void clamToSoftwareEnstops(float[] target)
  {

    if (target[X_AXIS] < X_MIN_PRT) target[X_AXIS] = X_MIN_PRT;
    if (target[Y_AXIS] < Y_MIN_PRT) target[Y_AXIS] = Y_MIN_PRT;
    if (target[Z_AXIS] < Z_MIN_PRT) target[Z_AXIS] = Z_MIN_PRT;

    if (target[X_AXIS] > X_MAX_PRT) target[X_AXIS] = X_MAX_PRT;
    if (target[Y_AXIS] > Y_MAX_PRT) target[Y_AXIS] = Y_MAX_PRT;
    if (target[Z_AXIS] > Z_MAX_PRT) target[Z_AXIS] = Z_MAX_PRT;
  }

  /** 
   ** scale() - scales the object
   **/
  public void scaleObject(int delta)
  {
    if (delta == -1) {
      scaler += 0.05;
    } // smaller
    if (delta == 1) {
      scaler -= 0.05;
    } // bigger
    if (mousePressed) {
      scaler = 1;
    } // reset scale

    scaler = constrain(scaler, .2, 8.0);
  }

  /** 
   ** moveObject() - scales the object
   **/
  public void moveObject(int x, int y)
  {
    nmx = x-xOffset; 
    nmy = y-yOffset;
  }

  /** 
   ** rotateObject() - rotates the object
   **/
  public void rotateObject(int x, int y)
  {
    rz -= (x - ejxOffset)*0.0005;
    if (rz > TWO_PI) { 
      rz = 0.0;
    }
    rx -= (y - ejyOffset)*0.0005;
    if (rx > TWO_PI) { 
      rx = 0.0;
    }
  }
  /** 
   ** setOffsetPoint() - sets the mouse offsets points
   **/
  public void setOffsetPoint(int x, int y)
  {
    xOffset = x-nmx; 
    yOffset = y-nmy;
    ejxOffset = x; 
    ejyOffset = y;
  }

  /**
   ** initObjectPosition() - Init the position of object
   **/
  void initObjectPosition() {
    if (activeTool == LASER_TOOL) {
      if (laserModeCtrl.getState(0)) { // If laser tool in plotting
        rx = .5;
        ry = 0;
        rz = 0;
        scaler = .7;
      } else if (laserModeCtrl.getState(1)) { // If laser tool in 3D
        rx = 1.3;
        ry = 0;
        rz = .8;
        scaler = .7;
      } else { // Laser in rastering
        rx = .5;
        ry = 0;
        rz = 0;
        scaler = .2;
      }
    } else {  // Rest of tools
      rx = 1.3;
      ry = 0;
      rz = .8;
      scaler = .7;
    }
    nmx =  nmy =  nmz = 0;
  }
  /**
   ** centerObject() - Center the object on the drawing window
   **/
  private void centerObject() {
    nmx = DRAW_WINDOW_SIZE[0]/2;
    nmy = DRAW_WINDOW_SIZE[1]/2;
    nmz = DRAW_WINDOW_SIZE[2]/2;
  }


  /**
   ** cleanDrawWindow() - Clean the drawing window
   **/
  void cleanDrawWindow() {
    mypg.beginDraw();
    mypg.background(G3DDBACKGROUND);  // Clean the window
    mypg.endDraw();
  }

  ///////////// UTILITY methods /////////////



  /**
   * hypot() - Calc hypotenuse
   **/
  public float hypot(float cat1, float cat2) {

    return sqrt(cat1*cat1+cat2*cat2);
  }

  /**
   ** codeSeen() 
   **/
  boolean codeSeen(char code)
  {
    char c;

    while ( (c = cmdbuffer[buffinx++]) != '\0' && buffinx < MAX_CMD_SIZE) {
      if (c == code)
        return true;
    }
    return false;
  }



  /**
   ** codeValue() - only accepts float positive numbers ATENCION!!!!!
   ** 
   **/
  float codeValue()throws NumberFormatException 
  {
    char c;
    int n = 0;
    char[] numbuf =  new char[MAX_FLOAT_SIZE+1];

    while (buffinx < MAX_CMD_SIZE) {
      c = cmdbuffer[buffinx++];
      if (((c >= '0') && (c <= '9')) || (c == '.') || (c == '+') || (c == '-')) {
        numbuf[n++] = c;
        if (n > MAX_FLOAT_SIZE)
          break;
      } else { // end of number
        numbuf[n] = '\0';

        return (float)atof(numbuf);
      }
    }
    throw new NumberFormatException ("Error parsing float number");
  }

  /**
   * atof() - Processing doesn´t have.
   **/
  float atof( char[] s) { 
    String str2 = new String(s);

    return float(str2);
  }
}